import rubxy

from typing import Optional

from .object import Object

class Contact(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        phone_number: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None
    ):
        super().__init__(client=client)
        
        self.phone_number = phone_number
        self.first_name = first_name
        self.last_name = last_name


import rubxy

from rubxy import types, enums, utils
from typing import Optional, Union

class EditMessageKeypad:
    async def edit_message_keypad(
        self: "rubxy.Client",
        chat_id: Union[str, int],
        message_id: Union[str, int],
        chat_keypad: Optional["types.Keypad"] = None,
        inline_keypad: Optional["types.Keypad"] = None,
        chat_keypad_type: Optional["enums.ChatKeypadType"] = enums.ChatKeypadType.NONE
    ):
        chat_keypad, inline_keypad, chat_keypad_type = utils.keypad_parse(chat_keypad, inline_keypad, chat_keypad_type)

        r = await self.invoke(
            "editMessageKeypad",
            chat_id=chat_id,
            message_id=message_id,
            chat_keypad=chat_keypad,
            inline_keypad=inline_keypad,
            chat_keypad_type=chat_keypad_type
        )

        return types.Message(
            chat_id=chat_id,
            id=message_id,
            client=self
        )


import rubxy

from rubxy import types, utils
from typing import Union, Optional

class EditMessageText:
    async def edit_message_text(
        self: "rubxy.Client",
        chat_id: Union[str, int],
        text: str,
        message_id: Union[str, int],
    ) -> "types.Message":
        r = await self.invoke(
            "editMessageText",
            text=text,
            chat_id=chat_id,
            message_id=message_id,
        )

        return types.Message(
            text=text,
            chat_id=chat_id,
            id=message_id,
            client=self
        )


import rubxy

class getUpdates:
    async def get_updates(
        self: "rubxy.Client",
        limit: int = 100,
        offset_id: str = None
    ):
        return await self.invoke(
            method="getUpdates",
            offset_id=offset_id,
            limit=limit
        )


from enum import auto
from .auto_name import AutoName

class ChatKeypadType(AutoName):
    NONE = auto()
    "None type, default"

    NEW = auto()
    "Add new keypad"

    REMOVE = auto()
    "Remove keypad"


import rubxy

from rubxy import enums
from typing import Optional

from .object import Object

class Chat(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        chat_id: Optional[str] = None,
        chat_type: Optional["enums.ChatType"] = None,
        user_id: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        title: Optional[str] = None,
        username: Optional[str] = None
    ):
        super().__init__(client=client)
        
        self.chat_id = chat_id
        self.chat_type = chat_type
        self.user_id = user_id
        self.first_name = first_name
        self.last_name = last_name
        self.title = title
        self.username = username


from rubxy import enums
from typing import Optional

from .object import DictLike

class ButtonTextbox(DictLike):
    def __init__(
        self,
        type_line: Optional["enums.ButtonTextboxTypeLine"] = None,
        type_keypad: Optional[enums.ButtonTextboxTypeKeypad] = None,
        place_holder: Optional[str] = None,
        title: Optional[str] = None,
        default_value: Optional[str] = None
    ):
        self.type_line = type_line
        self.type_keypad = type_keypad
        self.place_holder = place_holder
        self.title = title
        self.default_value = default_value


from enum import auto
from .auto_name import AutoName

class ButtonCalendarType(AutoName):
    DATE_PERSIAN = auto()
    "Display calendar in Persian format"

    DATE_GREGORIAN = auto()
    "Display calendar in Gregorian format"


import rubxy

from rubxy import enums, types

class UpdateEndpoints:
    async def update_endpoints(
        self: "rubxy.Client",
        url: str,
        type: "enums.UpdateEndpointType"
    ):
        return await self.invoke(
            "updateBotEndpoints",
            url=url,
            type=type
        )


import rubxy

from rubxy.errors import raise_exception
from typing import Optional

HEADERS = {'Content-Type': 'application/json'}

class Invoke:
    async def invoke(
        self: "rubxy.Client",
        method: str,
        **parameters,
    ) -> dict:
        if not self.is_started:
            raise ConnectionError("Client has not been started yet")
        
        await self._rate_limit()
        
        async with self.http.post(method, json=parameters, headers=HEADERS) as response:
            response.raise_for_status() # raise exception if not response.ok
            result = await response.json()

            if result.get("status") != "OK":
                raise_exception(
                    result.get("status"),
                    result.get("dev_message")
                )
            
            return result.get("data")


import rubxy

from rubxy import enums
from typing import Optional

from .object import DictLike

class ButtonCalendar(DictLike):
    def __init__(
        self,
        default_value: Optional[str] = None,
        type: Optional["enums.ButtonCalendarType"] = None,
        min_year: Optional[str] = None,
        max_year: Optional[str] = None,
        title: Optional[str] = None
    ):
        self.default_value = default_value
        self.type = type
        self.min_year = min_year
        self.max_year = max_year
        self.title = title


import typing
import json
import rubxy

from datetime import datetime
from enum import Enum

class Object:
    def __init__(self, client: "rubxy.Client" = None):
        self._client = client
    
    @staticmethod
    def default(obj: "Object"):
        if isinstance(obj, bytes):
            return repr(obj)
        
        if isinstance(obj, typing.Match):
            return repr(obj)

        if isinstance(obj, Enum):
            return str(obj)

        if isinstance(obj, datetime):
            return str(obj)
        
        attributes_to_hide = ()

        filtered_attributes = {
            attr: getattr(obj, attr)
            for attr in filter(
                lambda x: not x.startswith("_") and x not in attributes_to_hide,
                obj.__dict__,
            )
            if getattr(obj, attr) is not None
        }

        return dict(**filtered_attributes)
    
    def __str__(self) -> str:
        return json.dumps(self, indent=4, default=Object.default, ensure_ascii=False)
    
    def __getattr__(self, name):
        return None

class DictLike:
    def __repr__(self):
        return f"rubxy.types.{self.__class__.__name__}"
        
    def _to_dict(self):
        return self.__dict__


import rubxy

from rubxy import enums, types
from typing import Union, Optional, Match

from .object import Object

class Update(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        type: Optional["enums.UpdateType"] = None,
        chat_id: Optional[str] = None,
        removed_message_id: Optional[str] = None,
        new_message: Optional["types.Message"] = None,
        updated_message: Optional["types.Message"] = None,
        updated_payment: Optional["types.PaymentStatus"] = None,
        matches: Optional[list[Match]] = None,
    ):
        super().__init__(client=client)
        
        self.type = type
        self.chat_id = chat_id
        self.removed_message_id = removed_message_id
        self.new_message = new_message
        self.updated_message = updated_message
        self.updated_payment = updated_payment
        self.matches = matches

    async def reply(
        self: Union["types.Update", "types.Message", "types.InlineMessage"],
        text: str,
        metadata: Optional["types.MetaData"] = None,
        qoute: Optional[bool] = None,
        chat_id: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        chat_keypad: Optional["types.Keypad"] = None,
        inline_keypad: Optional["types.Keypad"] = None,
        reply_to_message_id: Optional[str] = None,
        chat_keypad_type: Optional["enums.ChatKeypadType"] = enums.ChatKeypadType.NONE
    ):
        if qoute is None:
            qoute = not self.chat_id.startswith(('u', 'b')) # true if message in group
        
        if qoute and reply_to_message_id is None:
            reply_to_message_id = self.message_id or self.new_message.id or self.updated_message.id

        return await self._client.send_message(
            text=text,
            metadata=metadata,
            chat_id=chat_id or self.chat_id,
            disable_notification=disable_notification,
            chat_keypad=chat_keypad,
            inline_keypad=inline_keypad,
            reply_to_message_id=reply_to_message_id,
            chat_keypad_type=chat_keypad_type
        )
    
    async def delete(
        self: Union["types.Update", "types.Message", "types.InlineMessage"],
        chat_id: Optional[Union[str, int]] = None,
        message_id: Optional[Union[str, int]] = None
    ):
        return await self._client.delete_message(
            chat_id=chat_id or self.chat_id,
            message_id=message_id or self.id or self.new_message.id or self.updated_message.id
        )
    
    def continue_propagation(self):
        raise rubxy.ContinuePropagation
    
    def stop_propagation(self):
        raise rubxy.StopPropagation


import rubxy

from rubxy import enums, types
from typing import Optional, List, Match

from .update import Update
from .object import Object

class Message(Update, Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        chat_id: Optional[str] = None,
        id: Optional[str] = None,
        text: Optional[str] = None,
        time: Optional[int] = None,
        is_edited: Optional[bool] = None,
        sender_type: Optional["enums.SenderType"] = None,
        sender_id: Optional[str] = None,
        aux_data: Optional["types.AuxData"] = None,
        file: Optional["types.File"] = None,
        reply_to_message_id: Optional[str] = None,
        forwarded_from: Optional["types.ForwardedFrom"] = None,
        matches: Optional[List[Match]] = None,
        metadata: Optional["types.MetaData"] = None
    ):
        super().__init__(client=client)
        
        self.chat_id = chat_id
        self.id = id
        self.text = text
        self.time = time
        self.is_edited = is_edited
        self.sender_type = sender_type
        self.sender_id = sender_id
        self.aux_data = aux_data
        self.file = file
        self.reply_to_message_id = reply_to_message_id
        self.forwarded_from = forwarded_from
        self.matches = matches
        self.metadata = metadata



import rubxy

from rubxy import types, enums, utils
from filetype import guess_mime

from typing import Union, Optional

class SendFile:
    async def send_file(
        self: "rubxy.Client",
        chat_id: Union[str, int],
        file_id: Union[str, int] = None,
        file: Union[str, bytes] = None,
        file_name: Optional[str] = None,
        file_type: Optional["enums.FileType"] = None,
        text: str = None,
        reply_to_message_id: Optional[int] = None,
        disable_notification: Optional[bool] = False,
        chat_keypad: Optional["types.Keypad"] = None,
        inline_keypad: Optional["types.Keypad"] = None,
        chat_keypad_type: Optional["enums.ChatKeypadType"] = enums.ChatKeypadType.NONE
    ):
        chat_keypad, inline_keypad, chat_keypad_type = utils.keypad_parse(chat_keypad, inline_keypad, chat_keypad_type)
        
        if not file_id:
            if isinstance(file, str):
                file = open(file, "rb")
            elif isinstance(file, bytes):
                pass
            else:
                raise TypeError("`file` must be of type bytes or str, not %s"%type(file))
            
            if file_type is None:
                guess = guess_mime(file)
                file_type = guess.split('/')[0].capitalize()

                if file_type not in list(enums.FileType):
                    raise TypeError(f"filetype not allowed, allowed types: {list(enums.FileType)}")

            upload_url = await self.request_send_file(file_type)
            file_id = await self.upload_file(upload_url, file, file_name)

        r = await self.invoke(
            "sendFile",
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad=chat_keypad,
            inline_keypad=inline_keypad,
            chat_keypad_type=chat_keypad_type
        )

        return types.Message(
            chat_id=chat_id,
            text=text,
            id=r.get("message_id"),
            reply_to_message_id=reply_to_message_id,
            client=self
        )


import re
import inspect
import asyncio
import rubxy

from rubxy import enums
from rubxy.types import Update, InlineMessage
from rubxy.utils import anies
from typing import Callable, Any, Union, Optional

class Filter:
    async def __call__(self, client, update) -> bool:
        raise NotImplementedError

    def __and__(self, other: "Filter") -> "Filter":
        return AndFilter(self, other)

    def __or__(self, other: "Filter") -> "Filter":
        return OrFilter(self, other)

    def __invert__(self) -> "Filter":
        return InvertFilter(self)


class InvertFilter(Filter):
    def __init__(self, base: Filter):
        self.base = base

    async def __call__(self, client: "rubxy.Client", update) -> bool:
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(client, update)
        else:
            x = await client.loop.run_in_executor(client.executor, self.base, client, update)
        return not x


class AndFilter(Filter):
    def __init__(self, base: Filter, other: Filter):
        self.base = base
        self.other = other

    async def __call__(self, client: "rubxy.Client", update) -> bool:
        # base
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(client, update)
        else:
            x = await client.loop.run_in_executor(client.executor, self.base, client, update)

        if not x:  # short-circuit
            return False

        # other
        if inspect.iscoroutinefunction(self.other.__call__):
            y = await self.other(client, update)
        else:
            y = await client.loop.run_in_executor(client.executor, self.other, client, update)

        return x and y


class OrFilter(Filter):
    def __init__(self, base: Filter, other: Filter):
        self.base = base
        self.other = other

    async def __call__(self, client: "rubxy.Client", update: Update) -> bool:
        # base
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(client, update)
        else:
            x = await client.loop.run_in_executor(client.executor, self.base, client, update)

        if x:  # short-circuit
            return True

        # other
        if inspect.iscoroutinefunction(self.other.__call__):
            y = await self.other(client, update)
        else:
            y = await client.loop.run_in_executor(client.executor, self.other, client, update)

        return x or y

class commands(Filter):
    def __init__(self, commands: Union[list, str], prefixes: Optional[list] = ['/'], ignore_case: bool = False):
        self.ignore_case = ignore_case
        cmds = commands if isinstance(commands, list) else [commands]
        escaped_prefixes = [re.escape(p) for p in prefixes]

        self.pattern = r"^(?:{})(?:{})".format(
            "|".join(escaped_prefixes),
            "|".join(cmds)
        ) + r"(?:\s|$)"
    
    def __call__(self, _, update: Update):
        _text = update.new_message.text or update.updated_message.text
        
        if not _text:
            return False
        
        p = re.compile(self.pattern, re.IGNORECASE if self.ignore_case else 0)
        update.matches = list(p.finditer(_text))
            
        return bool(update.matches)


class regex(Filter):
    def __init__(self, pattern: str, flags: int = 0):
        self.p = re.compile(pattern, flags)

    def __call__(self, _, update: Union[Update, InlineMessage]):
        string = (
            getattr(update.new_message, "text", None) or getattr(update.updated_message, "text", None)
        ) or (
            getattr(update.aux_data, "button_id", None) or update.text
        )
        
        if not string:
            return False
        
        update.matches = list(self.p.finditer(string))

        return bool(update.matches)
    
class private_filter(Filter):
    def __call__(self, _, update: Update):
        return update.chat_id.startswith(("u", "b"))
    
private = private_filter()

class group_filter(Filter):
    def __call__(self, _, update: Update):
        return update.chat_id.startswith("g0")

group = group_filter()

class channel_filter(Filter):
    def __call__(self, _, update: Update) -> bool:
        return update.chat_id.startswith("c0")

channel = channel_filter()

class text_filter(Filter):
    def __call__(self, _, update: Update) -> bool:
        return anies(
            update.new_message.text,
            update.updated_message.text
        )

text = text_filter()

class file_filter(Filter):
    def __call__(self, _, update: Update) -> bool:
        return anies(
            update.new_message.file,
            update.updated_message.file
        )

file = file_filter()

class forwarded_filter(Filter):
    def __call__(self, _, update: Update):
        return anies(
            update.new_message.forwarded_from,
            update.updated_message.forwarded_from
        )
    
forwarded = forwarded_filter()

class chat(Filter):
    def __init__(self, chat_id: Union[list[str, int], str]):
        self.chat_id = chat_id

        if isinstance(self.chat_id, str):
            self.chat_id = [chat_id]

        elif isinstance(self.chat_id, list): pass
        else:
            raise TypeError("`chat_id` must be type of list or str")
    
    async def __call__(self, _, update: Update):
        return update.chat_id in self.chat_id

class edited_filter(Filter):
    async def __call__(self, _, update: Update):
        return getattr(update.new_message, "is_edited", None) or getattr(update.new_message, "is_edited", None)
    
edited = edited_filter()


import rubxy

from rubxy import enums
from typing import Optional

from .object import Object

class PaymentStatus(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        payment_id: Optional[str] = None,
        status: Optional["enums.PaymentStatus"] = None
    ):
        super().__init__(client=client)

        self.payment_id = payment_id
        self.status = status


import rubxy
import inspect
import traceback

from rubxy.types import Update
from rubxy.filters import Filter
from typing import Callable

class Handler:
    def __init__(
        self,
        callback: Callable,
        filters: Filter = None,
        event_type: str = None
    ):
        self.callback = callback
        self.filters = filters
        self.event_type = event_type
    
    async def check(
        self,
        client: "rubxy.Client",
        update: Update
    ):
        if not self.filters:
            await self._call_callback(client, update)           
            return True

        # All filters must be true
        f = self.filters
        
        try:
            if inspect.iscoroutinefunction(f.__call__):
                ok = await f(client, update)
            else:
                ok = f(client, update)
        except Exception:
            ok = False

        if not ok:
            return
        
        await self._call_callback(client, update)
        return True

    async def _call_callback(
        self,
        client: "rubxy.Client",
        update: Update
    ):
        if inspect.iscoroutinefunction(self.callback):
            await self.callback(client, update)
        else:
            client.loop.run_in_executor(
                client.executor,
                self.callback,
                client, update
            )


import rubxy

from rubxy import enums
from rubxy.filters import Filter
from typing import Optional, Callable

from .handler import Handler

class StartHandler(Handler):
    def __init__(
        self,
        callback: Callable,
    ):
        super().__init__(callback=callback)


from enum import Enum

class AutoName(str, Enum):
    def _generate_next_value_(self, *args):
        return ''.join(word.capitalize() for word in self.split('_'))
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)
    
    def __str__(self):
        return self.value
    
    def __repr__(self):
        return f"'{self.value}'"
    
    def __hash__(self):
        return hash(self.value)


import logging
import warnings
import asyncio
import time
import inspect
import os
import sys

from aiohttp import ClientSession, ClientTimeout, TCPConnector
from concurrent.futures import ThreadPoolExecutor
from collections import deque

from rubxy import enums
from rubxy.methods import Methods
from rubxy.handlers import Handler
from rubxy.dispatcher import Dispatcher
from rubxy.parser import Markdown

from typing import Any, Union, Literal, Dict, List, Callable, Optional
from importlib import import_module
from pathlib import Path

from ._config import (
    DEAFULT_HOST,
    DEAFULT_PORT,
    DEAFULT_PATH,
    DEAFULT_TIMEOUT,
    DEAFULT_RATE_LIMIT,
    DEAFULT_POLL_INTERVAL
)

logger = logging.getLogger(__name__)

class Client(Methods):
    """
    Rubika Client, the main means for interacting with Rubika.

    ---

    **Parameters:**

    - `bot_token` (`str`): Pass the Bot API token to create a bot.
    - `plugins` (`dict`, optional): Smart Plugins settings as dict, e.g. `dict(root="plugins")`.
    - `timeout` (`float`, optional): Long polling wait time in seconds (default: 30).
    """

    def __init__(
        self,
        bot_token: str,
        plugins: Optional[dict] = None,
        timeout: Optional[float] = DEAFULT_TIMEOUT,
        rate_limit: Optional[float] = DEAFULT_RATE_LIMIT,
        poll_interval: Optional[float] = DEAFULT_POLL_INTERVAL,
        base_url: Optional[str] = None,
        parse_mode: Optional[Union[str, enums.ParseMode]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        loop: Optional[asyncio.BaseEventLoop] = None
    ):
        self.bot_token: str = bot_token
        self.base_url: str = (base_url or "https://botapi.rubika.ir/v3/{}/").format(bot_token)
        self.http: ClientSession = None
        self.connector: TCPConnector = None
        self._formatter = Markdown(self)
        self.is_started: bool = False
        self.is_long_polling: bool = None
        self.middlewares: List[Callable] = []
        self.dispatcher: Dispatcher = Dispatcher(self, poll_interval=poll_interval)
        self.plugins: dict = plugins
        self.timeout: int = timeout
        self.parse_mode = parse_mode
        self.completed_updates = deque(maxlen=200)
        self.rate_limit: float = rate_limit
        self.last_request = 0
        self.executor = executor or ThreadPoolExecutor(min(32, (os.cpu_count() or 1) + 4))
        self.loop = loop

        self.start_handlers: List[Callable] = []
        self.stop_handlers: List[Callable] = []
        
        if isinstance(loop, asyncio.AbstractEventLoop):
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()

        # parse_mode normalizer

        if self.parse_mode is None:
            self.parse_mode = enums.ParseMode.MARKDOWN

        elif isinstance(self.parse_mode, str):
            if self.parse_mode.lower() in ("mk", "markdown"):
                self.parse_mode = enums.ParseMode.MARKDOWN
            elif self.parse_mode.lower() in ("html"):
                self.parse_mode = enums.ParseMode.HTML
            else:
                raise ValueError("parse_mode must be type of enums.ParseMode or in [html, mk, markdown]")
        
    async def start(self):
        if not self.is_started:
            self.is_started = True
            timeout = ClientTimeout(total=self.timeout)
            
            if self.connector is None:
                self.connector = TCPConnector(limit=50)
            
            if self.http is None:
                self.http = ClientSession(
                    base_url=self.base_url,
                    connector=self.connector,
                    timeout=timeout,
                )
        
        for handler in self.start_handlers:
            if inspect.iscoroutinefunction(handler):
                await handler(self)
            else:
                self.loop.run_in_executor(
                    self.executor,
                    handler,
                    self
                )
        
        return self
    
    async def stop(self):
        if self.is_started:
            self.is_started = False
            self.dispatcher.is_running = False

            for handler in self.stop_handlers:
                if inspect.iscoroutinefunction(handler):
                    await handler(self)
                else:
                    self.loop.run_in_executor(
                        self.executor,
                        handler,
                        self
                    )

    def __enter__(self):
        self.start()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    async def __aenter__(self):
        await self.start()
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.stop()

    async def _rate_limit(self):
        passed = time.time() - self.last_request

        if passed < self.rate_limit:
            await asyncio.sleep(self.rate_limit - passed)
        
        self.last_request = time.time()

    def load_plugins(self):
        if not self.plugins:
            return
        
        plugins = self.plugins.copy()
        root = plugins.get("root")
        include = plugins.get("include", [])
        exclude = plugins.get("exclude", [])
        count = 0

        if not root or not isinstance(root, str):
            raise ValueError("`root` parameter is required for plugins")

        for path in sorted(Path(root.replace('.', '/')).rglob('*.py')):
            module_path = '.'.join(path.parent.parts + (path.stem,))
            logger.info("`%s` module loading...", module_path)
            module = import_module(module_path)

            for name in vars(module).keys():
                try:
                    for handler, group in getattr(module, name).handlers:
                        self.add_handler(handler, group)
                        count += 1
                except Exception:
                    pass
        
        logger.info("%s plugins are loaded successfuly", count)

    def run(
        self,
        endpoint: Optional[str] = None,
        host: Optional[str] = DEAFULT_HOST,
        path: Optional[str] = DEAFULT_PATH,
        port: Optional[int] = DEAFULT_PORT,
        update_endpoints: Optional[bool] = True
    ) -> None:
        self.load_plugins()

        runner = self.loop.run_until_complete
        runner(self.start())

        if endpoint:
            if endpoint.startswith('http'):
                runner(self.dispatcher.setup_webhook(host=host, path=path, port=port, endpoint=endpoint, update_endpoints=update_endpoints))

                try:
                    self.loop.run_forever()
                except KeyboardInterrupt as e:
                    runner(self.stop())
                    raise e
                finally:
                    runner(self.dispatcher.app.shutdown())
                    runner(self.dispatcher.app.cleanup())
            else:
                raise ValueError("Endpoint format is incorrect, it should be starts with http")
        else:
            warnings.warn(
                "Note: long-polling cannot receive `InlineMessage` updates",
                UserWarning
            )

            self.is_long_polling = True

            try:
                runner(self.dispatcher.updater(limit=100))
            except KeyboardInterrupt as e:
                runner(self.stop())
                raise e


from enum import auto
from .auto_name import AutoName

class ForwardedFrom(AutoName):
    USER = auto()
    "Forward from user"

    CHANNEL = auto()
    "Forward from channel"

    BOT = auto()
    "Forward from bot"


from .on_start import OnStart
from .on_stop import OnStop
from .on_message import OnMessage
from .on_inline_message import OnInlineMessage

class Decorators(
    OnStart,
    OnStop,
    OnMessage,
    OnInlineMessage
):
    pass


import rubxy

from rubxy import enums
from typing import Optional, List

from .object import Object

class PollStatus(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        state: Optional["enums.PollStatus"] = None,
        selection_index: Optional[int] = None,
        percent_vote_options: Optional[List[int]] = None,
        total_vote: Optional[int] = None,
        show_total_votes: Optional[bool] = None
    ):
        super().__init__(client=client)
        
        self.state = state
        self.selection_index = selection_index
        self.percent_vote_options = percent_vote_options
        self.total_vote = total_vote
        self.show_total_vote = show_total_votes


from .bot_command import BotCommand
from .message import Message
from .update import Update
from .inline_message import InlineMessage
from .keypad import Keypad
from .chat import Chat
from .bot import Bot
from .file import File
from .aux_data import AuxData
from .forwarded_from import ForwardedFrom
from .payment_status import PaymentStatus
from .button_selection import ButtonSelection
from .button_selection_item import ButtonSelectionItem
from .button_calendar import ButtonCalendar
from .button_number_picker import  ButtonNumberPicker
from .button_location import ButtonLocation
from .button_textbox import ButtonTextbox
from .button_string_picker import ButtonStringPicker
from .button_link import ButtonLink
from .open_chat_data import OpenChatData
from .join_channel_data import JoinChannelData
from .button import Button
from .location import Location
from .poll_status import PollStatus
from .metadata import MetaData
from .meta_data_parts import MetaDataParts
from .object import Object


import rubxy
import aiohttp

from random import randint

class UploadFile:
    async def upload_file(
        self: "rubxy.Client",
        upload_url,
        file,
        file_name: str = None
    ):
        if file_name is None:
            file_name = "file-{}".format(randint(1, 999999))

        form = aiohttp.FormData()
        form.add_field(
            name="file",
            value=file,
            filename=file_name,
            content_type="application/octet-stream"
        )

        async with self.http.post(upload_url, data=form) as response:
            response.raise_for_status()
            result = await response.json()
            
            return result["data"]["file_id"]


import rubxy

from rubxy import enums, types
from typing import Optional, List

from .object import DictLike

class ButtonSelection(DictLike):
    def __init__(
        self,
        selection_id: Optional[str] = None,
        search_type: Optional[str] = None,
        get_type: Optional[str] = None,
        items: Optional[List["types.ButtonSelectionItem"]] = None,
        is_multi_selection: Optional[bool] = None,
        columns_count: Optional[str] = None,
        title: Optional[str] = None
    ):
        self.selection_id = selection_id
        self.search_type = search_type
        self.get_type = get_type
        self.items = items
        self.is_multi_selection = is_multi_selection
        self.columns_count = columns_count
        self.title = title


import rubxy

from typing import Optional

from .object import DictLike

class ButtonNumberPicker(DictLike):
    def __init__(
        self,
        min_value: Optional[str] = None,
        max_value: Optional[str] = None,
        default_value: Optional[str] = None,
        title: Optional[str] = None
    ):
        self.min_value = min_value
        self.max_value = max_value
        self.default_value = default_value
        self.title = title



from enum import auto
from .auto_name import AutoName

class ButtonSelectionSearch(AutoName):
    NONE = auto()
    "Default state"

    LOCAL = auto()
    "Search list items using the values sent in the items field"

    API = auto()
    "Search list items via API"


import rubxy

from rubxy import types
from typing import Union

class GetChat:
    async def get_chat(
        self: "rubxy.Client",
        chat_id: Union[str, int]
    ) -> "types.Chat":
        r = await self.invoke(
            "getChat",
            chat_id=chat_id
        )

        chat = r.get("chat", {})

        return types.Chat(
            chat_id=chat.get("chat_id"),
            chat_type=chat.get("chat_type"),
            user_id=chat.get("user_id"),
            first_name=chat.get("first_name"),
            last_name=chat.get("last_name"),
            title=chat.get("title"),
            username=chat.get("username"),
            client=self
        )


from enum import auto
from .auto_name import AutoName

class UpdateType(AutoName):
    UPDATED_MESSAGE = auto()
    "Update message, like edit message"

    NEW_MESSAGE = auto()
    "New message"

    REMOVED_MESSAGE = auto()
    "Remove message"

    STARTED_BOT = auto()
    "Started bot"

    STOPPED_BOT = auto()
    "Stopped bot"

    UPDATED_PAYMENT = auto()
    "Updated payment"


from enum import auto
from .auto_name import AutoName

class ButtonSelectionType(AutoName):
    TEXT_ONLY = auto()
    "Display button as text"

    TEXT_IMG_THU = auto()
    "Display button as text and thumbnail"

    TEXT_IMG_BIG = auto()
    "Display the button as text and large image"


import rubxy

from rubxy import types
from typing import Optional

from .object import Object

class Sticker(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        sticker_id: Optional[str] = None,
        file: Optional["types.File"] = None,
        emoji_character: Optional[str] = None
    ):
        super().__init__(client=client)

        self.sticker_id = sticker_id
        self.file = file
        self.emoji_character = emoji_character


import rubxy

from rubxy import types

from typing import Optional, List

from .object import DictLike

class MetaData(DictLike):
    def __init__(
        self,
        meta_data_parts: Optional[List["types.MetaDataParts"]],
    ):
        self.meta_data_parts = meta_data_parts


import rubxy

from rubxy import types
from typing import Optional

from .update import Update
from .object import Object

class InlineMessage(Update, Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        sender_id: Optional[str] = None,
        text: Optional[str] = None,
        file: Optional["types.File"] = None,
        location: Optional["types.Location"] = None,
        aux_data: Optional["types.AuxData"] = None,
        message_id: Optional[str] = None,
        chat_id: Optional[str] = None
    ):
        super().__init__(client=client)
        
        self.sender_id = sender_id
        self.text = text
        self.file = file
        self.location = location
        self.aux_data = aux_data
        self.message_id = message_id
        self.chat_id = chat_id


import rubxy

from rubxy import types
from typing import Optional

from .object import Object

class Bot(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        bot_id: Optional[str] = None,
        bot_title: Optional[str] = None,
        avatar: Optional["types.File"] = None,
        description: Optional[str] = None,
        username: Optional[str] = None,
        start_message: Optional[str] = None,
        share_url: Optional[str] = None,
    ):
        super().__init__(client=client)
        
        self.bot_id = bot_id
        self.bot_title = bot_title
        self.avatar = avatar
        self.description = description
        self.username = username
        self.start_message = start_message
        self.share_url = share_url


import rubxy

from rubxy import handlers

class OnStop:
    def on_stop(self: "rubxy.Client"):
        def decorator(func):
            self.add_handler(handlers.StopHandler(func))
            
            return func
        return decorator


import rubxy

from typing import Optional

from .object import Object

class File(Object):
    def __init__(
        self,
        *,
        client: Optional["rubxy.Client"] = None,
        file_id: Optional[str] = None,
        file_name: Optional[str] = None,
        size: Optional[str] = None
    ):
        super().__init__(client=client)

        self.file_id = file_id
        self.file_name = file_name
        self.size = size



from typing import Optional

from .object import DictLike

class MetaDataParts(DictLike):
    def __init__(
        self,
        from_index: Optional[int] = None,
        length: Optional[int] = None,
        type: Optional[str] = None,
        link_url: Optional[str] = None,
        mention_text_user_id: Optional[str] = None,
    ):
        self.from_index = from_index
        self.length = length
        self.type = type
        self.link_url = link_url
        self.mention_text_user_id = mention_text_user_id


import rubxy

from rubxy import types, enums, utils
from typing import Optional

class SendLocation:
    async def send_location(
        self: "rubxy.Client",
        chat_id: str,
        latitude: str,
        longitude: str,
        disable_notification: Optional[bool] = False,
        chat_keypad: Optional["types.Keypad"] = None,
        inline_keypad: Optional["types.Keypad"] = None,
        reply_to_message_id: Optional[int] = None,
        chat_keypad_type: Optional["enums.ChatKeypadType"] = enums.ChatKeypadType.NONE
    ) -> "types.Message":
        chat_keypad, inline_keypad, chat_keypad_type = utils.keypad_parse(chat_keypad, inline_keypad, chat_keypad_type)

        r = await self.invoke(
            "sendLocation",   
            chat_id=chat_id,
            latitude=latitude,
            longitude=longitude,
            disable_notification=disable_notification,
            chat_keypad=chat_keypad,
            inline_keypad=inline_keypad,
            reply_to_message_id=reply_to_message_id,
            chat_keypad_type=chat_keypad_type
        )

        return types.Message(
            chat_id=chat_id,
            id=r.get("message_id"),
            client=self
        )


import rubxy

from rubxy import types
from typing import Union, Optional

class ForwardMessage:
    async def forward_message(
        self: "rubxy.Client",
        from_chat_id: Union[str, int],
        message_id: Union[str, int],
        to_chat_id: Union[str, int],
        disable_notification: Optional[bool] = False
    ) -> "types.Message":
        r = await self.invoke(
            "forwardMessage",
            from_chat_id=from_chat_id,
            message_id=message_id,
            to_chat_id=to_chat_id,
            disable_notification=disable_notification
        )

        return types.Message(
            id=r.get("new_message_id"),
            client=self
        )


from rubxy import types, enums

def anies(*args):
    return any(args)

def keypad_parse(chat_keypad, inline_keypad, chat_keypad_type=None):
    if anies(chat_keypad, inline_keypad):
        if isinstance(chat_keypad, types.Keypad):
            chat_keypad = chat_keypad._to_dict()
            chat_keypad_type = enums.ChatKeypadType.NEW if chat_keypad_type is enums.ChatKeypadType.NONE else chat_keypad_type

        elif isinstance(inline_keypad, types.Keypad):
            inline_keypad = inline_keypad._to_dict()
        else:
            raise TypeError("`chat_keypad` or `inline_keypad` must be of type types.Keypad")
    
        return chat_keypad, inline_keypad, chat_keypad_type
    return chat_keypad, inline_keypad, chat_keypad_type


from enum import auto
from .auto_name import AutoName

class ButtonType(AutoName):
    SIMPLE = auto()
    "Display the button normally"

    SELECTION = auto()
    "Display the button as a list"

    CALENDAR = auto()
    "Display the button as a calendar"

    NUMBER_PICKER = auto()
    "Display the button as a list of numbers"

    STRING_PICKER = auto()
    "Display the button as a list of strings"

    LOCATION = auto()
    "Location type"

    PAYMENT = auto()
    "Display the button for payment"

    CAMERA_IMAGE = auto()
    "Display the button for taking a photo with the camera"

    CAMERA_VIDEO = auto()
    "Display the button for recording a video with the camera"

    GALLERY_IMAGE = auto()
    "Show button to send image from gallery"

    GALLERY_VIDEO = auto()
    "Show button to send video from gallery"

    FILE = auto()
    "Show button to send a file"

    AUDIO = auto()
    "Show button to send audio"

    RECORD_AUDIO = auto()
    "Show button to record audio"

    MY_PHONE_NUMBER = auto()
    "My phone number"

    MY_LOCATION = auto()
    "My location"

    TEXTBOX = auto()
    "Show button to enter text message"

    LINK = auto()
    "Show button to send internet address"

    ASK_MY_PHONE_NUMBER = auto()
    "Request my phone number"

    ASK_LOCATION = auto()
    "Request location"

    BARCODE = auto()
    "Show button to scan barcode"


import rubxy

from rubxy import types

class getMe:
    async def get_me(
        self: "rubxy.Client",
    ) -> "types.Bot":
        r = (await self.invoke("getMe")).get("bot")

        return types.Bot(
            bot_id=r.get("bot_id", ""),
            bot_title=r.get("bot_title", ""),
            description=r.get("description", ""),
            username=r.get("username", ""),
            start_message=r.get("start_message", ""),
            share_url=r.get("share_url", ""),
            client=self
        )


from .advanced import Advanced
from .chats import Chats
from .decorators import Decorators
from .messages import Messages
from .updates import Updates
from .users import Users
from .utilities import Utilities
from .settings import Settings

class Methods(
    Advanced,
    Chats,
    Decorators,
    Messages,
    Updates,
    Users,
    Utilities,
    Settings
):
    pass


from rubxy import types
from typing import Optional

from .object import DictLike

class ButtonLink(DictLike):
    def __init__(
        self,
        link_url: Optional[str] = None,
        joinchannel_data: Optional["types.JoinChannelData"] = None,
        open_chat_data: Optional["types.OpenChatData"] = None
    ):
        self.link_url = link_url
        self.joinchannel_data = joinchannel_data
        self.open_chat_data = open_chat_data


from typing import Optional
from rubxy import enums, types

from .object import DictLike

class Button(DictLike):
    def __init__(
        self,
        id: Optional[str] = None,
        type: Optional[enums.ButtonType] = enums.ButtonType.SIMPLE,
        button_text: Optional[str] = None,
        button_selection: Optional["types.ButtonSelection"] = None,
        button_calendar: Optional["types.ButtonCalendar"] = None,
        button_number_picker: Optional["types.ButtonNumberPicker"] = None,
        button_string_picker: Optional["types.ButtonStringPicker"] = None,
        button_location: Optional["types.ButtonLocation"] = None,
        button_textbox: Optional["types.ButtonTextbox"] = None,
    ):
        self.id = id
        self.type = type
        self.button_text = button_text
        self.button_selection = button_selection
        self.button_calendar = button_calendar
        self.button_number_picker = button_number_picker
        self.button_string_picker = button_string_picker
        self.button_location = button_location
        self.button_textbox = button_textbox


from .exceptions import (
    ApiError,
    InvalidAccess,
    InvalidInput,
    InvalidMethod,
    TooRequests
)

def raise_exception(
    status: str = None,
    dev_message: str = None
):
    _exceptions = {
        "INVALID_ACCESS": InvalidAccess,
        "INVALID_INPUT": InvalidInput,
        "INVALID_METHOD": InvalidMethod,
        "TOO_REQUESTS": TooRequests,
        "_": ApiError
    }

    _status = "_" if status is None or status not in _exceptions else status

    raise _exceptions[_status](status, dev_message)


import rubxy

from typing import Optional

from .object import Object

class AuxData(Object):
    def __init__(
        self,
        *, 
        client: "rubxy.Client" = None,
        start_id: Optional[str] = None,
        button_id: Optional[str] = None
    ):
        super().__init__(client=client)
        
        self.start_id = start_id
        self.button_id = button_id


class TooRequests(Exception):
    def __init__(self, status: str = None, dev_message: str = None):
        super().__init__(status, dev_message)

        self.status = status
        self.dev_message = dev_message
    
    def __str__(self):
        return (
            "Rubika API Exception: "
            "status: {} | dev-message: {}".format(self.status, self.dev_message)
        )


import rubxy

from rubxy import enums, types
from typing import Optional

from .object import DictLike

class ButtonLocation(DictLike):
    def __init__(
        self,
        default_pointer_location: Optional["types.Location"] = None,
        default_map_location: Optional["types.Location"] = None,
        type: Optional["enums.ButtonLocationType"] = None,
        title: Optional[str] = None,
        location_image_url: Optional[str] = None
    ):
        self.default_pointer_location = default_pointer_location
        self.default_map_location = default_map_location
        self.type = type
        self.title = title
        self.location_image_url = location_image_url


import rubxy

from rubxy import types
from typing import Union

class DeleteMessage:
    async def delete_message(
        self: "rubxy.Client",
        chat_id: Union[str, int],
        message_id: Union[str, int]
    ) -> "types.Message":
        r = await self.invoke(
            "deleteMessage",
            chat_id=chat_id,
            message_id=message_id
        )

        return types.Message(
            chat_id=chat_id,
            id=message_id,
            client=self
        )


from typing import List, Optional
from rubxy import types

class Keypad:
    def __init__(
        self,
        rows: List[List["types.Button"]],
        resize_keyboard: Optional[bool] = None,
        on_time_keyboard: Optional[bool] = None
    ):
        self.rows = rows
        self.resize_keyboard = resize_keyboard
        self.on_time_keyboard = on_time_keyboard
    
    def _to_dict(self):
        return dict(
            rows=[
                dict(
                    buttons=[
                        button._to_dict() for button in row
                    ]
                ) for row in self.rows
            ],
            resize_keyboard=self.resize_keyboard,
            on_time_keyboard=self.on_time_keyboard
        )
    


import rubxy
import asyncio
import logging
import inspect
import time

from rubxy import types, enums, handlers, errors
from collections import OrderedDict
from typing import Callable, Dict, List, Union
from asyncio import sleep
from aiohttp import web, ClientResponseError
from aiohttp.web import Application, Request

logger = logging.getLogger(__name__)

ALLOWED_IPS: list = []

class Dispatcher:
    def __init__(
        self,
        client: "rubxy.Client",
        poll_interval: float
    ):
        self.client = client
        self.application: Application = None 
        self.groups = {}
        self.last_offset = None
        self.poll_interval = poll_interval
        self.is_running = False
        self.is_first_update = True
        self.lock = asyncio.Lock()

    def has_time_passed(
        self,
        last_time,
        seconds
    ):
        try:
            timestamp = int(float(last_time))
            now = time.time()
            return (now - timestamp) > seconds
        except (TypeError, ValueError):
            return False
    
    async def add_handler(
        self,
        handler: "handlers.Handler",
        group: int = 0
    ):
        group = int(group)

        async with self.lock:
            event_type = handler.event_type
            
            if event_type not in self.groups:
                self.groups[event_type] = OrderedDict()

            if group not in self.groups[event_type]:
                self.groups[event_type][group] = []

            self.groups[event_type][group].append(handler)
            self.groups[event_type] = OrderedDict(sorted(self.groups[event_type].items()))
    
    async def updater(self, limit: int):
        self.is_running = True
        self.next_offset_id = None

        try:
            while self.is_running:
                try:
                    updates = await self.client.invoke(
                        "getUpdates",
                        offset_id=self.next_offset_id,
                        limit=limit
                    )
                    
                    if not all(
                        [
                            updates.get("updates"),
                            updates,
                        ]
                    ):
                        await sleep(self.poll_interval)
                        continue
                    
                    self.next_offset_id = updates.get("next_offset_id", self.next_offset_id)
                    updates = updates.get("updates", [])

                    if self.is_first_update:
                        updates = []
                        self.is_first_update = False
                        continue
                    
                    for raw_update in updates:
                        last_time = raw_update.get('new_message', {}).get('time') or raw_update.get('updated_message', {}).get('time')
                        
                        if self.has_time_passed(last_time, 10):
                            break

                        asyncio.create_task(
                            self.process_update(
                                raw_update,
                                enums.EventType.INLINE_MESSAGE if 'inline_message' in raw_update else enums.EventType.MESSAGE
                            )
                        )
                
                except ClientResponseError:
                    if self.client.is_long_polling:
                        pass
                    
                except Exception:
                    logger.exception('an error occurred while polling update')
        except KeyboardInterrupt:
            await self.client.stop()

    async def process_update(self, updates: Union[Dict, List[Dict]], event_type: enums.EventType):
        if isinstance(updates, dict):
            updates = [updates]
                
        for raw_update in updates:
            update = None

            if event_type is enums.EventType.MESSAGE:
                raw_update = raw_update.get("update", raw_update) or {}
                new_message = raw_update.get("new_message") or {}
                aux_data = new_message.get("aux_data") or {}
                file = new_message.get("file") or {}
                forwarded_from = new_message.get("forwarded_from") or {}
                updated_message = raw_update.get("updated_message") or {}
                updated_payment = raw_update.get("updated_payment") or {}
                updated_message_aux_data = updated_message.get("aux_data") or {}
                updated_message_file = updated_message.get("file") or {}
                updated_message_forwarded_from = updated_message.get("forwarded_from") or {}

                update = types.Update(
                    type=raw_update.get("type"),
                    chat_id=raw_update.get("chat_id"),
                    removed_message_id=raw_update.get("removed_message_id"),
                    new_message=types.Message(
                        id=new_message.get("message_id"),
                        text=new_message.get("text"),
                        time=new_message.get("time"),
                        is_edited=new_message.get("is_edited"),
                        sender_type=new_message.get("sender_type"),
                        sender_id=new_message.get("sender_id"),
                        aux_data=types.AuxData(
                            start_id=aux_data.get("start_id"),
                            button_id=aux_data.get("button_id"),
                            client=self.client
                        ) if aux_data else None,
                        file=types.File(
                            file_id=file.get("file_id"),
                            file_name=file.get("file_name"),
                            size=file.get("size"),
                            client=self.client
                        ) if file else None,
                        reply_to_message_id=new_message.get("reply_to_message_id"),
                        forwarded_from=types.ForwardedFrom(
                            type_from=forwarded_from.get("type_from"),
                            id=forwarded_from.get("message_id"),
                            from_chat_id=forwarded_from.get("from_chat_id"),
                            from_sender_id=forwarded_from.get("from_sender_id"),
                            client=self.client
                        ) if forwarded_from else None,
                        metadata=new_message.get("metadata"),
                        client=self.client
                    ) if new_message else None,
                    updated_message=types.Message(
                        id=updated_message.get("message_id"),
                        text=updated_message.get("text"),
                        time=updated_message.get("time"),
                        is_edited=updated_message.get("is_edited"),
                        sender_type=updated_message.get("sender_type"),
                        sender_id=updated_message.get("sender_id"),
                        aux_data=types.AuxData(
                            start_id=updated_message_aux_data.get("start_id"),
                            button_id=updated_message_aux_data.get("button_id"),
                            client=self.client
                        ) if updated_message_aux_data else None,
                        file=types.File(
                            file_id=updated_message_file.get("file_id"),
                            file_name=updated_message_file.get("file_name"),
                            size=updated_message_file.get("size"),
                            client=self.client
                        ) if updated_message_file else None,
                        reply_to_message_id=updated_message.get("reply_to_message_id"),
                        forwarded_from=types.ForwardedFrom(
                            type_from=updated_message_forwarded_from.get("type_from"),
                            id=updated_message_forwarded_from.get("message_id"),
                            from_chat_id=updated_message_forwarded_from.get("from_chat_id"),
                            from_sender_id=updated_message_forwarded_from.get("from_sender_id"),
                            client=self.client
                        ) if updated_message_forwarded_from else None,
                        metadata=updated_message.get("metadata"),
                    ) if updated_message else None,
                    updated_payment=types.PaymentStatus(
                        payment_id=updated_payment.get("payment_id"),
                        status=updated_payment.get("status"),
                        client=self.client
                    ) if updated_payment else None,
                    client=self.client
                )

            elif event_type is enums.EventType.INLINE_MESSAGE:
                raw_update = raw_update.get("inline_message") or {}
                file = raw_update.get("file") or {}
                location = raw_update.get("location") or {}
                aux_data = raw_update.get("aux_data") or {}

                update = types.InlineMessage(
                    sender_id=raw_update.get("sender_id"),
                    text=raw_update.get("text"),
                    file=types.File(
                        file_id=file.get("file_id"),
                        file_name=file.get("file_name"),
                        size=file.get("size"),
                        client=self.client
                    ) if file else None,
                    location=types.Location(
                        longitude=location.get("longitude"),
                        latitude=location.get("latitude"),
                        client=self.client
                    ) if location else None,
                    aux_data=types.AuxData(
                        start_id=aux_data.get("start_id"),
                        button_id=aux_data.get("button_id"),
                        client=self.client
                    ) if aux_data else None,
                    message_id=raw_update.get("message_id"),
                    chat_id=raw_update.get("chat_id"),
                    client=self.client
                )
            
            await self.dispatch_update(update=update, event_type=event_type)

    async def dispatch_update(self, update, event_type):
        async def run_middlewares(index: int):
            if index < len(self.client.middlewares):
                mw = self.client.middlewares[index]

                if inspect.iscoroutinefunction(mw):
                    await mw(self.client, update, lambda: run_middlewares(index + 1))
                else:
                    self.client.loop.run_in_executor(
                        self.client.executor,
                        mw,
                        self.client, update, lambda: self.client.loop.create_task(run_middlewares(index + 1))
                    )
            else:
                event_handlers: Dict[int, List[handlers.Handler]] = self.groups.get(event_type) or {}

                for group, _handlers in event_handlers.items():
                    for _handler in _handlers:
                        try:
                            _check = await _handler.check(self.client, update)

                            if _check:
                                break
                            
                        except rubxy.StopPropagation:
                            break
                        except rubxy.ContinuePropagation:
                            continue
        
        await run_middlewares(0)
    
    def handle_webhook(self, event_type: enums.EventType):
        async def handle(request: Request):
            data = await request.json()
            
            asyncio.create_task(self.process_update(data, event_type=event_type))

            return web.json_response(
                {"status": "OK"},
                status=200
            )

        return handle
    
    def get_real_ip(self, request: Request):
        if 'X-Forwarded-For' in request.headers:
            return request.headers['X-Forwarded-For'].split(',')[0].strip()
        return request.remote
    
    @web.middleware
    async def ip_range_check(self, request: Request, handler: Callable):
        ip = self.get_real_ip(request)

        if ip not in ALLOWED_IPS:
            return web.json_response(
                {'error': 'Access denied'},
                status=403
            )
        
        return await handler(request)
    
    async def setup_webhook(self, host: str, path: str, port: int, endpoint: str, update_endpoints: bool):
        _path = path.rstrip('/')
        pather = lambda a=None, b=_path: f"{b}/{a}" if a else ""
        decapitalize = lambda s: s[:1].lower() + s[1:] if s else s
        handler = self.handle_webhook

        self.app = Application()
        self.app.router.add_post(pather(), handler(enums.EventType.MESSAGE))
        self.app.router.add_post(pather("receiveUpdate"), handler(enums.EventType.MESSAGE))
        self.app.router.add_post(pather("receiveInlineMessage"), handler(enums.EventType.INLINE_MESSAGE))

        if update_endpoints:
            for type in list(enums.UpdateEndpointType):
                _endpoint = endpoint.rstrip('/') + path
                
                result = await self.client.update_endpoints(
                    url=pather(
                        a=decapitalize(type),
                        b=_endpoint
                    ),
                    type=type
                )

                if result.get("status") != "Done":
                    raise errors.ApiError(status=result.get("status"))

                logger.info(
                    'Endpoint: %s | Type: %s | Status: %s',
                    _endpoint, type, result.get("status", "failure").lower()
                )

        runner = web.AppRunner(self.app)
        await runner.setup()

        site = web.TCPSite(runner, host=host, port=port)
        await site.start()


import rubxy

from rubxy import types
from typing import List

class SendPoll:
    async def send_poll(
        self: "rubxy.Client",
        chat_id: str,
        question: str,
        options: List[str]
    ) -> "types.Message":
        r = await self.invoke(
            "sendPoll",
            chat_id=chat_id,
            question=question,
            options=options
        )

        return types.Message(
            chat_id=chat_id,
            id=r.get("message_id"),
            client=self
        )


from .send_message import SendMessage
from .send_contact import SendContact
from .send_location import SendLocation
from .send_poll import SendPoll
from .edit_message import EditMessage
from .edit_message_text import EditMessageText
from .edit_message_keypad import EditMessageKeypad
from .delete_message import DeleteMessage
from .forward_message import ForwardMessage
from .request_send_file import RequestSendFile
from .get_file import GetFile
from .send_file import SendFile
from .upload_file import UploadFile

class Messages(
    SendMessage,
    SendContact,
    SendLocation,
    SendPoll,
    DeleteMessage,
    EditMessage,
    EditMessageText,
    EditMessageKeypad,
    ForwardMessage,
    RequestSendFile,
    GetFile,
    SendFile,
    UploadFile
):
    pass


from rubxy import enums
from typing import Optional

from .object import DictLike

class OpenChatData(DictLike):
    def __init__(
        self,
        object_guid: Optional[str] = None,
        object_type: Optional["enums.ChatType"] = None
    ):
        self.object_guid = object_guid
        self.object_type = object_type


import rubxy

from rubxy import handlers

class OnStart:
    def on_start(self: "rubxy.Client"):
        def decorator(func):
            self.add_handler(handlers.StartHandler(func))
            
            return func
        return decorator


import rubxy

from typing import Optional, List

from .object import DictLike

class ButtonStringPicker(DictLike):
    def __init__(
        self,
        items: Optional[List[str]] = None,
        default_value: Optional[str] = None,
        title: Optional[str] = None
    ):
        self.items = items
        self.default_value = default_value
        self.title = title


import rubxy

from typing import Union

class GetFile:
    async def get_file(
        self: "rubxy.Client",
        file_id: Union[str, int]
    ):
        r = await self.invoke(
            "getFile",
            file_id=file_id
        )

        return r.get("download_url")


from enum import auto
from .auto_name import AutoName

class ChatType(AutoName):
    USER = auto()
    "Chat is a private chat with a user"

    BOT = auto()
    "Chat is a private chat with a bot"

    GROUP = auto()
    "Chat is a basic group"

    CHANNEL = auto()
    "Chat is the channel"


import rubxy

from rubxy import types
from typing import Optional, List

from .object import Object

class Poll(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        question: Optional[str] = None,
        options: Optional[List[str]] = None,
        poll_status: Optional["types.PollStatus"] = None
    ):
        super().__init__(client=client)
        
        self.question = question
        self.options = options
        self.poll_status = poll_status


from .message_handler import MessageHandler
from .inline_message_handler import InlineMessageHandler
from .start_handler import StartHandler
from .stop_handler import StopHandler
from .handler import Handler

class Handlers(
    MessageHandler,
    InlineMessageHandler,
    StartHandler,
    StopHandler
):
    pass


import rubxy

from rubxy import enums
from typing import Optional

from .object import DictLike

class ButtonSelectionItem(DictLike):
    def __init__(
        self,
        text: Optional[str] = None,
        image_url: Optional[str] = None,
        type: Optional["enums.ButtonSelectionType"] = None,
    ):
        self.text = text
        self.image_url = image_url
        self.type = type


import rubxy

from rubxy import types, enums
from typing import Optional

from .object import Object

class LiveLocation(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        start_time: Optional[str] = None,
        live_period: Optional[str] = None,
        current_location: Optional["types.Location"] = None,
        user_id: Optional[str] = None,
        status: Optional["enums.LiveLocationStatus"] = None,
        last_update_time: Optional[str] = None
    ):
        super().__init__(client=client)

        self.start_time = start_time
        self.live_period = live_period
        self.current_location = current_location
        self.user_id = user_id
        self.status = status
        self.last_update_time = last_update_time


import rubxy
import asyncio

from rubxy import handlers

class AddHandler:
    def add_handler(
        self: "rubxy.Client",
        handler: "handlers.Handler",
        group: int = 0
    ):
        if isinstance(handler, handlers.StartHandler):
            self.start_handlers.append(handler.callback)
        
        elif isinstance(handler, handlers.StopHandler):
            self.stop_handlers.append(handler.callback)
            
        else:
            self.loop.create_task(self.dispatcher.add_handler(handler, group))


import rubxy

from rubxy import enums
from rubxy.filters import Filter
from typing import Optional, Callable

from .handler import Handler

class MessageHandler(Handler):
    def __init__(
        self,
        callback: Callable,
        filters: Optional[Filter] = None,
    ):
        super().__init__(
            callback=callback,
            filters=filters,
            event_type=enums.EventType.MESSAGE
        )


from enum import auto
from .auto_name import AutoName

class ButtonTextboxTypeLine(AutoName):
    SINGLE_LINE = auto()
    "Write text message in a single line"

    MULTI_LINE = auto()
    "Write text message in multiple lines"


import asyncio

from typing import Optional, List

class Plugins:
    def __init__(
        self,
        root: str,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None
    ):
        self.root = root
        self.include = [include] if isinstance(include, str) else include
        self.exclude = [exclude] if isinstance(exclude, str) else exclude

class StopPropagation(StopAsyncIteration):
    pass

class ContinuePropagation(StopAsyncIteration):
    pass

try:
    asyncio.get_running_loop()
except RuntimeError:
    asyncio.set_event_loop(asyncio.new_event_loop())


from .client import Client
from . import sync


from typing import Optional

from .object import DictLike

class JoinChannelData(DictLike):
    def __init__(
        self,
        username: Optional[str] = None,
        ask_join: Optional[bool] = False
    ):
        self.username = username.replace("@", "") if username else username
        self.ask_join = ask_join        


import rubxy
import warnings

from rubxy import handlers
from rubxy.filters import Filter
from typing import Optional, Union

class OnInlineMessage:
    def on_inline_message(
        self: Union["rubxy.Client", "Filter"] = None,
        filters: Optional[Filter] = None,
        group: int = 0
    ):
        def decorator(func):
            if isinstance(self, rubxy.Client):
                self.add_handler(
                    handlers.InlineMessageHandler(func, filters),
                    group=group
                )
            
            elif isinstance(self, Filter) or self is None:
                if not hasattr(func, "handlers"):
                    func.handlers = []
                
                func.handlers.append(
                    (
                        handlers.InlineMessageHandler(func, self),
                        group if filters is None else filters
                    )
                )
            
            return func
        return decorator


import rubxy

from rubxy import enums
from typing import Optional

from .object import Object

class ForwardedFrom(Object):
    def __init__(
        self,
        *,
        client: "rubxy.Client" = None,
        type_from: Optional["enums.ForwardedFrom"] = None,
        id: Optional[str] = None,
        from_chat_id: Optional[str] = None,
        from_sender_id: Optional[str] = None
    ):
        super().__init__(client=client)

        self.type_from = type_from
        self.id = id
        self.from_chat_id = from_chat_id
        self.from_sender_id = from_sender_id


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="rubxy",
    author="Amirali Rahimi",
    author_email="amiralirahimi769@gmail.com",
    version="1.0.3",
    description="A Python library for interacting with the Rubika bot API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", )),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=["aiohttp", "filetype"],
    license="GPL-3.0-or-later",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "rubika",
        "rubxy",
        "bot",
        "robot"
    ],
)


import rubxy

from rubxy import types, enums, utils
from typing import Optional, Union

class EditMessage:
    async def edit_message(
        self: "rubxy.Client",
        chat_id: Union[str, int],
        text: str,
        message_id: Union[str, int],
        chat_keypad: Optional["types.Keypad"] = None,
        inline_keypad: Optional["types.Keypad"] = None,
        chat_keypad_type: Optional["enums.ChatKeypadType"] = enums.ChatKeypadType.NONE
    ) -> "types.Message":
        r = await self.edit_message_text(
            text=text,
            chat_id=chat_id,
            message_id=message_id
        )

        if utils.anies(chat_keypad, inline_keypad):
            if inline_keypad:
                await self.edit_message_keypad(
                    chat_id=r.chat_id,
                    message_id=r.id,
                    chat_keypad=chat_keypad,
                    inline_keypad=inline_keypad,
                    chat_keypad_type=chat_keypad_type
                )
        
        return r


from enum import auto
from .auto_name import AutoName

class ButtonSelectionGet(AutoName):
    LOCAL = auto()
    "Display list items using the values sent in the items field"

    API = auto()
    "Search list items via API"


import rubxy

from rubxy import types
from typing import List, Union

class SetCommands:
    async def set_commands(
        self: "rubxy.Client",
        bot_commands: Union["types.BotCommand", List["types.BotCommand"]]
    ):
        if isinstance(bot_commands, types.BotCommand):
            bot_commands = [bot_commands]
        
        bot_commands = [
            commands._to_dict() for commands in bot_commands
        ]

        r = await self.invoke(
            "setCommands",
            bot_commands=bot_commands
        )

        print(r)


from typing import Optional

from .object import DictLike

class BotCommand(DictLike):
    def __init__(
        self,
        command: Optional[str] = None,
        description: Optional[str] = None
    ):
        self.command = command.strip('/') if command else command
        self.description = description


from enum import auto
from .auto_name import AutoName

class ButtonTextboxTypeKeypad(AutoName):
    STRING = auto()
    "Allows sending all characters"

    NUMBER = auto()
    "Allows sending numeric characters"


from enum import auto
from .auto_name import AutoName

class UpdateEndpointType(AutoName):
    RECEIVE_UPDATE = auto()
    RECEIVE_INLINE_MESSAGE = auto()
    RECEIVE_QUERY = auto()
    GET_SELECTION_ITEM = auto()
    SEARCH_SELECTION_ITEMS = auto()


import rubxy

from rubxy import enums

class RequestSendFile:
    async def request_send_file(
        self: "rubxy.Client",
        type: enums.FileType
    ):
        if not isinstance(type, enums.FileType) and type not in list(enums.FileType):
            raise TypeError("`type` must be of type enums.FileType or in %s"%list(enums.FileType))
        
        r = await self.invoke(
            "requestSendFile",
            type=type
        )

        return r.get("upload_url")


from .button_selection_type import ButtonSelectionType
from .button_calendar_type import ButtonCalendarType
from .button_location_type import ButtonLocationType
from .button_selection_get import ButtonSelectionGet
from .button_selection_search import ButtonSelectionSearch
from .button_textbox_type_keypad import ButtonTextboxTypeKeypad
from .button_textbox_type_line import ButtonTextboxTypeLine
from .button_type import ButtonType
from .chat_keypad_type import ChatKeypadType
from .chat_type import ChatType
from .forwarded_from import ForwardedFrom
from .live_location_status import LiveLocationStatus
from .payment_status import PaymentStatus
from .poll_status import PollStatus
from .sender_type import SenderType
from .update_type import UpdateType
from .update_endpoint_type import UpdateEndpointType
from .event_type import EventType
from .file_type import FileType
from .parse_mode import ParseMode


# This module is based on the rubpy library.
# Used and modified with permission from the original author.
# Original source: https://github.com/shayanheidari01/rubika

import rubxy
import re
import markdownify

from rubxy import enums

from typing import Any, Dict, List

MARKDOWN_RE = re.compile(
    r"(?:^(?:> ?[^\n]*\n?)+)|```([\s\S]*?)```|\*\*([^\n*]+?)\*\*|`([^\n`]+?)`|__([^\n_]+?)__|--([^\n-]+?)--|~~([^\n~]+?)~~|\|\|([^\n|]+?)\|\||\[([^\]]+?)\]\((\S+)\)",
    flags=re.DOTALL | re.MULTILINE,
)

MARKDOWN_TYPES = {
    ">": ("Quote", None),
    "```": ("Pre", 1),
    "**": ("Bold", 2),
    "`": ("Mono", 3),
    "__": ("Italic", 4),
    "--": ("Underline", 5),
    "~~": ("Strike", 6),
    "||": ("Spoiler", 7),
    "[": ("Link", 8),
}

MENTION_PREFIX_TYPES = {"u": "User", "g": "Group", "c": "Channel", "b": "Bot"}

MARKDOWN_TYPE_SEQUENCE = tuple(MARKDOWN_TYPES.items())

def _build_utf16_prefix_lengths(text: str) -> List[int]:
    prefix_lengths = [0]
    total = 0
    append = prefix_lengths.append
    for char in text:
        total += 2 if ord(char) > 0xFFFF else 1
        append(total)
    return prefix_lengths


class Markdown:
    def __init__(self, client: "rubxy.Client"):
        self.client = client

    def _format_text(self, text):
        if self.client.parse_mode is enums.ParseMode.HTML:
            text = self.to_markdown(text)
        return self.to_metadata(text)
    
    def to_markdown(self, text: str) -> str:
        return markdownify.markdownify(html=text).strip()

    def to_metadata(self, text: str) -> Dict[str, Any]:
        meta_data_parts: List[Dict[str, Any]] = []
        current_text = text
        offset = 0
        char_offset = 0
        utf16_prefix = _build_utf16_prefix_lengths(text)

        for match in MARKDOWN_RE.finditer(text):
            group = match.group(0)
            start, end = match.span()
            adjusted_start = utf16_prefix[start] - offset
            adjusted_char_start = start - char_offset

            for prefix, (md_type, group_idx) in MARKDOWN_TYPE_SEQUENCE:
                if group.startswith(prefix):
                    if md_type == "Quote":
                        quote_lines = group.split('\n')
                        content_lines = []
                        for line in quote_lines:
                            if line.startswith('> '):
                                content_lines.append(line[2:])
                            elif line.startswith('>'):
                                content_lines.append(line[1:])
                            else:
                                content_lines.append(line)
                        content = '\n'.join(content_lines)
                        content_length = len(content.encode("utf-16-be")) // 2
                        char_content_length = len(content)
                        
                        inner_metadata = self.to_metadata(content)
                        content = inner_metadata["text"]
                        content_length = len(content.encode("utf-16-be")) // 2
                        char_content_length = len(content)
                        
                        if "metadata" in inner_metadata:
                            for part in inner_metadata["metadata"]["meta_data_parts"]:
                                part["from_index"] += adjusted_start
                                meta_data_parts.append(part)
                    else:
                        group_start = match.start(group_idx)
                        group_end = match.end(group_idx)
                        if group_start == -1 or group_end == -1:
                            content = ""
                            content_length = 0
                            char_content_length = 0
                        else:
                            content = match.group(group_idx) or ""
                            content_length = utf16_prefix[group_end] - utf16_prefix[group_start]
                            char_content_length = group_end - group_start
                            
                            if md_type not in ("Pre", "Link"):
                                inner_metadata = self.to_metadata(content)
                                content = inner_metadata["text"]
                                content_length = len(content.encode("utf-16-be")) // 2
                                char_content_length = len(content)
                                
                                if "metadata" in inner_metadata:
                                    for part in inner_metadata["metadata"]["meta_data_parts"]:
                                        part["from_index"] += adjusted_start
                                        meta_data_parts.append(part)
                    
                    meta_data_part = {
                        "type": md_type,
                        "from_index": adjusted_start,
                        "length": content_length,
                    }

                    if md_type == "Pre":
                        lines = content.split("\n", 1)
                        language = lines[0].strip() if lines[0].strip() else ""
                        meta_data_part["language"] = language
                    elif md_type == "Link":
                        url = match.group(9)
                        mention_type = MENTION_PREFIX_TYPES.get(url[0], "hyperlink")

                        if mention_type == "hyperlink":
                            meta_data_part["link_url"] = url
                            meta_data_part["link"] = {
                                "type": mention_type,
                                "hyperlink_data": {"url": url},
                            }
                        else:
                            meta_data_part["type"] = "MentionText"
                            meta_data_part["mention_text_object_guid"] = url
                            meta_data_part["mention_text_user_id"] = url
                            meta_data_part["mention_text_object_type"] = mention_type

                    meta_data_parts.append(meta_data_part)

                    markup_length = utf16_prefix[end] - utf16_prefix[start]
                    char_markup_length = end - start
                    current_text = (
                        current_text[:adjusted_char_start]
                        + content
                        + current_text[end - char_offset :]
                    )
                    offset += markup_length - content_length
                    char_offset += char_markup_length - char_content_length
                    break

        result = {"text": current_text.strip()}

        if meta_data_parts:
            result["metadata"] = {"meta_data_parts": meta_data_parts}

        return (
            current_text.strip(),
            {"meta_data_parts": meta_data_parts} if meta_data_parts else None
        )


import rubxy

from typing import Optional
from .object import Object

class Location(Object):
    def __init__(
        self,
        client: "rubxy.Client" = None,
        longitude: Optional[str] = None,
        latitude: Optional[str] = None
    ):
        super().__init__(client=client)
        
        self.longitude = longitude
        self.latitude = latitude    